import QueryBuilder from "./QueryBuilder";
const {
  openDatabase
} = require('react-native-sqlite-storage');

//SQLite.DEBUG(config.app.isDebug)

/**
 * Resultado de los querys
 */

/**
 * Administra la base de datos.
 */
export default class DB {
  static _instances = new Map();
  /**
   * Abre la base de datos
   * 
   * @param {string} name nombre de la db
   */
  constructor(name) {
    this.name = name;
  }
  static get(name) {
    var db = this._instances.get(name);
    if (!db) {
      db = new DB(name);
      this._instances.set(name, db);
    }
    return db;
  }

  /**
   * Abre la base de datos
   * @returns {SQLiteDatabase}
   */
  async open() {
    if (!this.sqlite) {
      //console.debug("DB.open", this.name)
      this.sqlite = await openDatabase({
        "name": this.name
      });
    }
    return this.sqlite;
  }

  /**
   * Cierra la db
   */
  async close() {
    if (this.sqlite) {
      this.sqlite.close();
      this.sqlite = null;
    }
  }

  /**
   * Ejecuta sentecias sql.
   * 
   * @param {string} sql sentencia
   * @param {any[]} params parametros
   * 
   * @returns {
   *  "insertId": $insertId, 
   *  "rows": [
   *   {"$columnName": $value}
   *  ]
   * }
   */
  executeSql(sql) {
    let params = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    return new Promise((txResolve, txReject) => {
      var _this$sqlite;
      (_this$sqlite = this.sqlite) === null || _this$sqlite === void 0 || _this$sqlite.transaction(tx => {
        new Promise((sqlResolve, sqlReject) => {
          tx.executeSql(sql, params, (_tx, result) => {
            sqlResolve({
              rows: result.rows.raw(),
              insertId: result.insertId,
              rowsAffected: result.rowsAffected
            });
          }, error => {
            sqlReject(error);
          });
        }).then(txResolve).catch(txReject);
      });
    });
  }

  /**
   * Obtiene el numero de version de la db
   */
  getVersion() {
    return this.executeSql("PRAGMA user_version").then(_ref => {
      let {
        rows
      } = _ref;
      return rows[0];
    }).then(row => row.user_version ?? 0);
  }

  /**
   * Establece la versión de la base de datos.
   * @param {int} version la nueva versión de la base de datos
   */
  setVersion(version) {
    return this.executeSql("PRAGMA user_version = " + version).then(_ref2 => {
      let {
        rows
      } = _ref2;
      return rows[0];
    });
  }

  /**
   * Inicializa el scheme de la base de datos.
   */
  async init(scheme, version) {
    console.debug("DB.init");
    await this.open();

    // Obtiene la version actual de la db
    const dbVersion = await this.getVersion();

    // Valida la version actual con la version del archivo
    if (dbVersion == 0) {
      await scheme.onCreate(this);
    } else if (dbVersion != version) {
      await scheme.onUpdate(this, dbVersion, version);
    }

    // Setea la nueva version en la base de datos
    this.setVersion(version);
    scheme.onLoad(this);
  }

  /** 
   * SELECT
   * 
   * @param {string} table name
   * @param {QueryOptions} options
   * @returns array de registros
   */
  query(table) {
    var _options$where;
    let options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    const sql = QueryBuilder.query(table, options);
    return this.executeSql(sql, (_options$where = options.where) === null || _options$where === void 0 ? void 0 : _options$where.args).then(result => result.rows);
  }

  /** 
   * INSERT
   * 
   * @param {string} table name
   * @param {any} values
   * @returns insertId
   */
  insert(table, values) {
    var sql = QueryBuilder.create(table, values);
    const params = Object.values(values);
    return this.executeSql(sql, params).then(result => result.insertId);
  }

  /** 
   * UPDATE
   * 
   * @param {string} table name
   * @param {any} values
   * @param {Where} where
   * @returns rowsAffected
   */
  update(table, values, where) {
    const sql = QueryBuilder.update(table, values, where.clause);
    const params = Object.values(values);
    const whereArgs = where.args ?? [];
    return this.executeSql(sql, [...params, ...whereArgs]).then(result => result.rowsAffected);
  }

  /** 
   * DELETE
   * 
   * @param {string} table name
   * @param {Where} where
   * @returns rowsAffected
   */
  delete(table, where) {
    const sql = QueryBuilder.destroy(table, where.clause);
    return this.executeSql(sql, where.args).then(result => result.rowsAffected);
  }

  /**
   * Compila una clausula where
   * 
   * @param {any[]} array [1,3,4,2]
   * @returns (?, ?, ?, ?)
   */
  static whereClausePlaceholders(array) {
    var args = [];
    for (let i = 0; i < array.length; i++) {
      args.push("?");
    }
    return "(" + args.join(", ") + ")";
  }

  /**
   * @param sqlString "valor con 'comillas'"
   * @returns "'valor con \'comillas\''"
   */
  static escapeSqlString(sqlString) {
    let escape = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '\'';
    var sb = [];
    sb.push(escape);
    if (sqlString.indexOf(escape) != -1) {
      var length = sqlString.length;
      for (var i = 0; i < length; i++) {
        var c = sqlString.charAt(i);
        if (c == escape) {
          sb.push(escape);
        }
        sb.push(c);
      }
    } else {
      sb.push(sqlString);
    }
    sb.push(escape);
    return sb.join("");
  }
}
//# sourceMappingURL=DB.js.map